#include "queue.h"

// General includes
#include <new>
using std::bad_alloc;

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;

#include <cassert>

namespace MyQueue {

    //! Construtor da fila com parâmetro.
    /*! Note que se nenhum valor for especificado (construtor padrão)
     *  o valor default é assumido (veja em queue.h).
     *  @param capacity Capacidade de armazenamento da fila.
     */
    template <class Object>
        Queue<Object>::Queue( int capacity ) :
            mCapacity( capacity ),
            mCurrentSize( 0 ),
            mFront( 0 ), // Inicia com zero pois incremento vem depois do dequeue()
            mBack( -1 ) // Inicia -1 pois primeiro incremente depois enfileira.
    {
        // Criar nova memória para armazenamento.
        try {  mStorageArea= new Object[ mCapacity ];  }
        catch ( const bad_alloc& e ) {
            // Aqui dentro vai a acao a ser tomada.
            cerr << "\n[Queue(int)]: Erro durante alocacao da área de armazenamento!\n";
            // Aborta o programa, usado apenas em testes.
            assert ( false );
        }
    }

    //! Construtor cópia.
    /*! @param clone Fila a ser copiada.
    */
    template <class Object>
        Queue<Object>::Queue( const Queue& clone ) :
            mCapacity( clone.mCapacity ),
            mCurrentSize( clone.mCurrentSize ),
            mFront( clone.mFront ),
            mBack( clone.mBack )
    {
        try {  mStorageArea = new Object[ mCapacity ];  }
        catch ( const bad_alloc& e ) {
            cerr << "\n[Queue( const Queue&)]: Erro durante alocacao da área de armazenamento!\n";
            assert ( false );
        }
        for ( int i = 0; i < mCapacity; ++i )
            mStorageArea[ i ] = clone.mStorageArea[ i ];
    }

    //! Destrutor padrão
    template <class Object>
        Queue<Object>::~Queue( ) { delete [] mStorageArea; }

    //! Operador de atribuição.
    /*! @param rhs Origem da atribuição.
    */
    template <class Object>
        const Queue<Object> & Queue<Object>::operator= ( const Queue & clone )
        {
            if ( this != &clone ) {
                for ( int i = 0; i < clone.mCapacity; ++i)
                    mStorageArea[ i ] = clone.mStorageArea[ i ];
                mCapacity    = clone.mCapacity;
                mCurrentSize = clone.mCurrentSize;
                mFront       = clone.mFront;
                mBack        = clone.mBack;
            }

            return *this;
        }


    //! Verifica se a fila está vazia.
    /*! A verificação da condição vazia é feita diretamente sobre o membro currentSize. 
     * @return Retorna verdadeiro se fila vazia, falso caso contrário.
     */
    template <class Object>
        bool Queue<Object>::isEmpty( ) const
        {
            return mCurrentSize == 0;
        }

    //! Retorna o elemento que está na frente da fila (sem removê-lo).
    /*! @return O elemento na frente da fila.
     *  @throw UnderflowException Se a fila estiver vazia.
     */
    template <class Object>
        const Object & Queue<Object>::front( ) const
        {
            if( isEmpty( ) )
                throw UnderflowException( );
            return mStorageArea[ mFront ];
        }

    //! Retorna o elemento que está no final da fila (sem removê-lo).
    /*! @return O elemento no final da fila.
     *  @throw UnderflowException Se a fila estiver vazia.
     */
    template <class Object>
        const Object & Queue<Object>::back( ) const
        {
            if( isEmpty( ) )
                throw UnderflowException( );
            return mStorageArea[ mBack ];
        }

    //! Retira o elemento da frente da fila, sem retorná-lo.
    /*! O comportamento "circular" da fila é garantido através
     *  do método increment()
     *  @throws UnderflowException se fila estiver fazia.
     *  @see increment
     */
    template <class Object>
        void Queue<Object>::dequeue( )
        {
            if( isEmpty( ) )
                throw UnderflowException( );

            mCurrentSize--;
            increment( mFront );
        }

    //! Insere um elemento no final da fila.
    /*! O comportamento "circular" da fila é garantido através
     *  do método increment()
     *  @throws OverflowException se fila estiver cheia.
     *  @see increment
     */
    template <class Object>
        void Queue<Object>::enqueue( const Object & x )
        {
            if( mCurrentSize == mCapacity )
                throw OverflowException( );

            increment( mBack );
            mStorageArea[ mBack ] = x;
            mCurrentSize++;
        }

    //! Limpa a fila, tornado-a vazia.
    template <class Object>
        void Queue<Object>::clear( )
        {
            mCurrentSize = 0;
            mFront = 0;
            mBack = -1;
        }

    //! Recupera a capacidade de armazenamento da fila.
    /*! @return A capacidade de armazenamento da fila.
     */
    template <class Object>
        int Queue<Object>::capacity( ) const 
        { return mCapacity; }

    //! Recupera a quantidade de elementos atualmente na fila.
    /*! @return A quantidade de elemento atualmente na fila.
     */
    template <class Object>
        int Queue<Object>::size( ) const 
        { return mCurrentSize; }

    //! Imprime a memória interna da fila. Para depuração apenas.
    template <class Object>
        void Queue<Object>::print( ) const 
        {
            int tempFront = mFront;

            cout << "[";
            for( int i = 0; i < mCurrentSize; ++i )
            {
                cout << mStorageArea[ tempFront ];
                increment( tempFront );
                if( i != mCurrentSize-1 ) cout << ',';
            }
            if ( mCurrentSize == 0 ) cout << " ";
            cout << "]";
        }

    //! Método interno para incrementar o índice de forma circular (do final para o início).
    /*! @param x O índice que será incrementado.
     */
    template <class Object>
        void Queue<Object>::increment( int & x ) const
        {
            if( ++x == mCapacity )
                x = 0;
        }
} // namespace MyQueue
